/* 
 * sdbInt.h --
 *
 *   This file describes the internal strucutres and functions
 *   of the simple database (sdb) extenstion. 
 *
 * Copyright (c) 1995  Tom Moore.  All rights reserved.
 *
 */

#ifndef _SDBINT
#define _SDBINT

#ifndef _TCLINT
#include "tcl.h"
#endif

#ifndef _SDB
#include "sdb.h"
#endif


/*----------------------------------------------------------------------
 *
 * Common include files
 *
 *----------------------------------------------------------------------
 */

#ifndef _TCL
#include <tcl.h>
#endif

#include <stdio.h>
#include <sys/time.h>
#include <string.h>
#include <malloc.h>

#include "bitstring.h"
#include "sdbPort.h"

/*----------------------------------------------------------------------
 *
 * Symbol and structure definition section 
 *
 *----------------------------------------------------------------------
 */

/* 
 * Maximum record length and maximum number of fields allowed in
 * a Simple Database ascii input file.
 */
#define SDB_RECORD_LENGTH 8192
#define SDB_MAX_FIELDS 500


/* 
 * These are the flags for variable types.  They may be or'ed
 * together in various combinations.
 */
#define SDB_CODE        1
#define SDB_NUMBER      (1<<1)
#define SDB_EXPRESSION  (1<<2)
#define SDB_CONSTANT    (1<<3)
#define SDB_OUTPUT      (1<<4)
#define SDB_SUMMARY     (1<<5)
#define SDB_ALLOCATED   (1<<6)
#define SDB_LOOKUP      (1<<7)
#define SDB_INDEXED     (1<<8)
#define SDB_MISSING     (1<<9)
#define SDB_MUTABLE     (1<<10)

/*
 * Error codes
 */
#define SDB_DIVZERO        1
#define SDB_MISSYIELD      (1<<1)
#define SDB_ASSIGNEXP      (1<<2)

/* 
 * The following defines the flavors of operator types
 * that can be generated by the parser.  Each operator
 * type may have different arguments.
 */

typedef enum {
    /* the first bunch of operator types evalutate to numeric values */
    constValue,
    unaryOp,
    binaryOp,
    booleanOp,
    simplexFnct,
    duplexFnct,
    hsgFnct,
    hsgNumber,
    hsgNumberIndex,
    numNumFnct,
    chrNumFnct,
    stringCompare,
    stringFind,
    numExpr,
    lookupNumber,
    ifNode,
    curveFnct,
    assignNum,
    charFnct,
    numSequence,
    randomFnct,

    /* the following are 'code' types.  Make sure that 'const_code'
     * is kept first in this list.  Other parts of the source code
     * will test for character expressions as any type >= 'const_code' */
    constCode,
    hsgCode,
    hsgCodeIndex,
    hsgCodeList,
    numChrFnct,
    chrChrFnct,
    charExpr,
    lookupCode,
    formatFnct,
    assignChar,
    charSequence,

    /*
     * The following is neither code nor numeric, but a cue
     * to be passed from the parser to the optimizer.  The
     * nodes below this part of the parse tree should not
     * be executed, but only provide parser information.
     * These nodes should be copied, but not searched
     * by the 'down_add' checker.
     */
    noScan
} SdbOperator;

/* 
 * The following define the op-codes for the 'binary_op' and
 * comparison operators.
 */
typedef enum {
    opSum,
    opSub,
    opMul,
    opDiv,
    opNeg,
    opGe,
    opGt,
    opLe,
    opLt,
    opEq,
    opNe,
    opAnd,
    opOr,
    opNop,
    opNot
} SdbOpIndex;


/* 
 * This is the structure that contains information about the 
 * ascii input database.  
 */

typedef struct SdbTextFile {
    enum {end, var, data, undefined} current; /* current record type */
    int                     split, /* split between code and numeric 
				    * variables in original segment */
                            k,   /* number of active columns in this segment */
                            m,   /* number of columns in this file segment */
                            n,   /* number of columns in inital file segment */
                            active,  /* flag if reading data */
                            itemsRead; /* number of columns this record */
    FILE                   *stream;
    char                   *list[SDB_MAX_FIELDS], 
                            record[SDB_RECORD_LENGTH], /* current record */
                          **listV;     /* pointer to structure allocated 
					* by Tcl_SplitList */
    int                     remap[SDB_MAX_FIELDS]; /* remap field locations if
						    * second VAR statement 
						    * encountered */
    char                   *field[SDB_MAX_FIELDS]; /* names of original 
						    * fields */
    int                     present[SDB_MAX_FIELDS];
    Tcl_Interp             *interp;
} SdbTextFile;

/* 
 * Set up a few forward references that are needed now but 
 * will be defined later on.
 */
struct SdbVariable;
struct SdbNode;

/* 
 * Strucuture that contains values added by a join table.
 * One structure is added for each record in table.
 */
typedef struct SdbTerminal {
    /* Make sure that the size of this structure is always a
     * multiple of 8 (the alignment requirement for doubles). 
     * Right now it is 2*4byte long (two pointers).  See comments
     * in the file 'join.c' for an explanation. */
    double                 *value;     /* pointer to array holding the numeric
					* values for this record */
    char                  **codes;     /* pointer to array holding the code 
					* values for this record */
} SdbTerminal;

/* 
 * A simple linked list to hold vertices in a yield curve.
 */
typedef struct SdbCurve {
    double                 *values;    /* pointer to vertix values */
    struct SdbCurve        *next;      /* pointer to next vertice */
} SdbCurve;

/* 
 * A structure to hold binary trees structures.  This is used for both
 * the join tables and the yield curves.
 */
typedef struct SdbTree {
    char                   *label;     /* the label associated with this node 
					* in the tree */
    void                   *tree;      /* The b-tree for the next level */
    struct SdbTree         *wild;      /* The b-tree for wild card matches, 
					* search this only if failed to match 
					* in the other tree */
    SdbTerminal            *terminal;  /* pointer to the record that matches 
					  this node */
} SdbTree;

/* 
 * List structures.  Used in 'in' function..
 */
typedef struct SdbList {
    int                     n;
    char                  **pntr;
} SdbList;

/* 
 * Random number streams.
 */
typedef struct SdbRandom {
    char                   *state;     /* state information required by 
					* setstate(3) */
    enum {uniform, normal} type;       /* types of random distributions
					* that are supported */
} SdbRandom;

/* 
 * Structure to contain information about lookup tables.  Only
 * one of these strucutres is created for each lookup table that
 * is read in.
 */
typedef struct SdbLookup {
    int                     referenceCount;  /* Number of times this table is 
					      * referenced (one count for each
					      * active variable).  Delete the 
					      * table when the reference count
                                              * reaches zero. */
    int                     numIndex,        /* Number of index variables */
                            numCodes,        /* # of added code variables */
                            numVals;         /* # of added numeric variables */
    char                  **indexNames,      /* names of index variables */
                          **codeNames,       /* names of code variables */
                          **valueNames;      /* names of value variables */
    SdbTree                *head;            /* Pointer to the first binary 
					      * structure that indexes this 
					      * join table. */
} SdbLookup;

/* 
 * linked list of expressions for conditionals 
 */
typedef struct SdbCat {
    struct SdbNode         *expr;
    SdbOpIndex              relation;
    struct SdbNode         *code;
    struct SdbCat          *next;
} SdbCat;

typedef struct SdbLookupHandle {
    SdbLookup              *luPtr;
    SdbTerminal            *cache;
    struct SdbVariable    **index;
    Tcl_HashTable          *remote;
    int                     referenceCount;
    unsigned int            stamp;
    char                   *name;
} SdbLookupHandle;

typedef struct SdbYieldSet {
    int                     numCats,
                            numValues;
    char                  **cats,
                          **values;
    SdbTree                *tree;
    struct SdbVariable    **list;
    unsigned int            stamp;
    SdbCurve               *cache;
} SdbYieldSet;

typedef struct SdbYieldHandle {
    SdbYieldSet            *yieldSet;  /* Pointer to a yield set */
    int                     xIndex,    /* Index of the independent variable */
                            yIndex;    /* Index of the dependent variable */
    struct SdbVariable    **list;      /* This will normally be NULL.  If
				        * set with a list of variable pointers,
				        * it will override the indexing in
				        * the main table.  (A way of supporting
				        * aliases.) */
    unsigned int            stamp;     /* cache counter */
    SdbCurve               *cache;
} SdbYieldHandle;

/* 
 * Linked list of names of variables.  Used to keep track of all 
 * the variables that reference or are referenced by the current variable.
 */
typedef struct SdbDependentList {
    char                   *id;        /* Variable name */
    struct SdbDependentList *next;     /* pointer to next item */
} SdbDependentList;

/* 
 * The following structure is the heart of the parser and
 * evaluator.  Each structure represents an operation and
 * its arguments.  Arguments may point to other structures
 * which may have operators and arguments.  The parser
 * builds this structure, the evaluator executes it. 
 */
typedef struct SdbNode {
    SdbOperator             type;      /* defines the operator for this node */
    struct SdbNode         *left;      /* standard arguments */
    struct SdbNode         *right;
    union {                            /* The following are specialized 
					* arguments that may vary in format 
					* depending on the operator. */
        char                   *code;  /* In the case of a 'const_code' node,
                                        *  pointer to a constant code value */
        int                     integer;  /* constant integer value */
        SdbOpIndex              op;    /* In the case of a binary or comparison
                                        * operator, the op-code. */
        double                  value; /* In the case of a 'const_value' 
				        * operator, the actual value. */
        struct SdbVariable     *varPtr; /* If this node refers to another 
					* variable, the pointer to that 
					* variable */
        double                  (*fnctptr) ();  /* If this node evaluates a 
					* function  (duplex_fnct, 
					* simplex_fnct, etc), then a pointer 
					* to that function. */
        SdbCat                 *classList;  /* linked list of expressions for 
                                        * conditionals (num_fnct). */
        SdbList                 list;  /* used in 'in' function */
        SdbLookupHandle        *luHandlePtr;  /* pointer to a lookup table 
					* handle. Used to define lookup table
					* variables */
        SdbYieldHandle         *yieldHandle;  /* Pointer and index into yield 
					* curve data.  Used for interpolate 
					* function. */
	SdbRandom              *randomPtr;  /* Pointer to random number 
					* generator data.  Used for random 
					* function */
    } content;
} SdbNode;

/* 
 * The following describes the structure that comprises the definition
 * of a variable in the database.  Each variable used one of these
 * strucutres.  The structures are stored as a linked list.  The order
 * of the columns in the database is altered by changing the order of
 * the linked list. 
 */
typedef struct SdbVariable {
    char                   *id;        /* The name of this variable */
    int                     flag;      /* A flag, made by or'ing the values
                                        * defined at the beginning of this 
					* file */
    unsigned int            stamp;     /* a timestamp, indicating the 
					*  freshenss of the cache value. */
    union {                            /* A cache of precomputed values.  
					* Values are valid if the time stamp 
					* is current. This should save some 
					* overhead during row operations when 
					* one variable refers to another. */
        char               *code;
        double              value;
      } cache;
    union {                            /* Here are the values that can be 
					* contained in this variable:  */
        int                 integer;   /* an integer constant (unused 
					* perhaps?) */
        double              value;     /* a numeric 'constant' (only one value
					* for the column, but can be changed 
					* at any time) */
        char               *code;      /* a character 'constant' (only one 
					* value for the column, but can be 
					* changed at any time) */
        char              **ivalue;    /* a column 'code' variable.  This is 
					* a pointer to an array, so that each 
					* row can have a unique value */
        double             *dvalue;    /* a column 'code' variable.  This is a
					* pointer to an array, so that each 
					* row can have a unique value */
        SdbNode            *expr[2];   /* Pointer to an expression.  There are
					* two copies of the expression: one in
					* original form and one after 
					* optimization. */
    } u;
    char                   *format;    /* a character string containing a 
					* printf format specifier for this 
					* variable */
    SdbDependentList       *up,
                           *down;      /* linked lists containing names of 
					* varaibles that this varaible is used
					* by ('up'), and variables that this 
					* variable uses ('down'). */
    struct SdbVariable     *next;      /* pointer to the next variable in the 
					* linked list . */
} SdbVariable;


/* 
 * A structure to define a stack of selection bitmaps.  The stack
 * is implemented as a linked list.  The selection bitmaps keep
 * track of the records that are selected.
 */
typedef struct SdbStack {
    _bit_type              *select;    /* Pointer to the selection bitmap. */
    struct SdbStack        *next;      /* Pointer to the next element in the 
					* stack */
} SdbStack;

/* 
 * A structure to pass as clientData into the variable trace
 * procedure.  When the trace is invoked the trace routine
 * will have access to the variable and database structures.
 */
typedef struct SdbItemData {
    SdbVariable            *varPtr;    /* Pointer to the variable */
    struct SdbFile         *dbPtr;     /* Pointer to the database */
} SdbItemData;

/* 
 * Structure to define a database.  One of these structures is
 * created for each database that is opened.  Contains lots of
 * important information:  pointer to the linked list of variables,
 * the current cache time stamp, record counts, selection stacks,
 * a hash table to contain all strings saved in the database, etc.
 */
typedef struct SdbFile {
    SdbVariable            *varPtr,
                           *indexPtr,
                          **outputPtrPtr;
    struct timeval          timePtr;
    int                     numRec,
                            position,
                            errors,
                            numItems,
                            varSize,
                            numSelect,
                            numOutput;
    unsigned int            stamp;
    char                   *dbname,
                           *filename;
    int                    *list,
                           *base,
                           *reverse;
    _bit_type              *select;
    SdbStack               *stack;
    Tcl_Interp             *interp;
    Tcl_HashTable          *htbl,
                           *yieldList,
                           *joinList,
                           *randomList;
    SdbItemData            *traceId;
} SdbFile;

/*----------------------------------------------------------------------
 *
 * Macro to simplify storing a string into a hash table, and returning
 * the location of the stored string.
 *
 *----------------------------------------------------------------------
 */
#define GetIndices(Htbl,code,new) \
        (Tcl_CreateHashEntry(Htbl, code, new))->key.string

/*----------------------------------------------------------------------
 *
 * global variables 
 *
 *----------------------------------------------------------------------
 */
EXTERN Tcl_HashTable    sdb_HashTable;



/*----------------------------------------------------------------------
 *
 * Procedure protos.  The following procedures are shared
 * between files, but are private to the Sdb package. 
 *
 *----------------------------------------------------------------------
 */

EXTERN void               SdbCloseDB _ANSI_ARGS_((SdbFile *sdbPtr));
EXTERN void               SdbRebuildOutput _ANSI_ARGS_((SdbFile *sdbPtr));
EXTERN double             SdbEvaluate _ANSI_ARGS_((SdbFile *, int,
			      SdbNode *));
EXTERN char *             SdbCharEval _ANSI_ARGS_((SdbFile *, int, SdbNode *));
EXTERN int                SdbSelect _ANSI_ARGS_((Tcl_Interp *, SdbFile *, 
			      int, char **));
EXTERN int                SdbPrettyPrint _ANSI_ARGS_((Tcl_Interp *, 
			      SdbVariable *, SdbNode *));
EXTERN int                SdbCalculate _ANSI_ARGS_((Tcl_Interp *, SdbFile *,
                              char *));
EXTERN int                SdbDefine _ANSI_ARGS_((Tcl_Interp *, SdbFile *, 
			      char *, char *));
EXTERN int                SdbDelete _ANSI_ARGS_((Tcl_Interp *, SdbFile *,
                              CONST char *, int fast));
EXTERN SdbVariable *      SdbFound _ANSI_ARGS_((SdbVariable *, const char *));
EXTERN char *             SdbWhatFnct _ANSI_ARGS_((double (*)()));
EXTERN SdbVariable *      SdbInsert _ANSI_ARGS_((SdbFile *, const char *, int,
                              int));
EXTERN int                SdbNameCheck _ANSI_ARGS_((char *));
EXTERN int                SdbReservedName _ANSI_ARGS_((char *));
EXTERN int                SdbJoinCmd _ANSI_ARGS_((Tcl_Interp *, SdbFile *,
                             int argc, char **argv));
EXTERN int                SdbSummary _ANSI_ARGS_((Tcl_Interp *, SdbFile *,
                              int argc, char **argv));
EXTERN void               SdbReOptimize _ANSI_ARGS_((SdbFile *sdbPtr));
EXTERN char *             SdbGetCode _ANSI_ARGS_((SdbFile *sdbPtr,
                              SdbVariable *varPtr, int rec));
EXTERN int                SdbSeek _ANSI_ARGS_((Tcl_Interp *, SdbFile *sdbPtr,
                              char *key));
EXTERN int                SdbIndex _ANSI_ARGS_((Tcl_Interp *, SdbFile *sdbPtr,
                              char *field));
EXTERN double             SdbGetNumber _ANSI_ARGS_((SdbFile *sdbPtr,
                              SdbVariable *varPtr, int rec));
EXTERN int                SdbArrange _ANSI_ARGS_((SdbFile *sdbPtr, int argc,
                               char **argv, int flag));
EXTERN int                SdbCloseFile _ANSI_ARGS_((Tcl_Interp *interp,
                              SdbFile *sdbPtr));
EXTERN SdbFile *          SdbOpenFile _ANSI_ARGS_((Tcl_Interp *interp,
                              char *string, char *dbname, SdbFile *sdbPtr));
EXTERN int                SdbReloadFile _ANSI_ARGS_((Tcl_Interp *interp,
                              SdbFile *SdbPtr, char *string));
EXTERN char *             SdbCodeLookup _ANSI_ARGS_((SdbFile *sdbPtr,
                              SdbLookupHandle *luPtr, int iNdex, int rec));
EXTERN double             SdbNumberLookup _ANSI_ARGS_((SdbFile *sdbPtr,
                              SdbLookupHandle *luPtr, int iNdex, int rec));
EXTERN void               SdbPutCodeLookup _ANSI_ARGS_((SdbFile *sdbPtr,
                              SdbLookupHandle *luPtr, int iNdex, int rec,
                              char *code));
EXTERN void               SdbPutNumberLookup _ANSI_ARGS_((SdbFile *sdbPtr,
                              SdbLookupHandle *luPtr, int iNdex, int rec,
                              double value));
EXTERN SdbVariable *      SdbInsertLookup _ANSI_ARGS_((SdbFile *sdbPtr,
                              CONST char *string, int flags,
                              SdbLookupHandle *luPtr, int iNdex));
EXTERN char *             SdbUpdate _ANSI_ARGS_((ClientData clientData,
                              Tcl_Interp *interp, char *name1,
                              char *name2, int flags));
EXTERN void               SdbDelJoinTree _ANSI_ARGS_((SdbTree *));
EXTERN double             SdbYieldLookup _ANSI_ARGS_((SdbFile *sdbPtr,
                              SdbYieldHandle *yieldHandle,
			      int rec, SdbNode *expr));
EXTERN void               SdbDeleteYieldSet _ANSI_ARGS_((
                              SdbYieldSet *yieldSet));
EXTERN int                SdbCurves _ANSI_ARGS_((Tcl_Interp *interp,
                              SdbFile *sdbPtr, int argc, char **argv));
EXTERN int                SdbSort _ANSI_ARGS_((Tcl_Interp *interp,
                              SdbFile *sdbPtr, int argc, char **argv));
EXTERN int                SdbRandomCmd _ANSI_ARGS_((Tcl_Interp *interp,
                              SdbFile *sdbPtr, int argc, char **argv));
EXTERN SdbRandom *        SdbCreateRandom _ANSI_ARGS_((SdbFile *sdbPtr,
                              char *sequence));
EXTERN SdbFile *          SdbCreateDB _ANSI_ARGS_((Tcl_Interp *interp,
                              char *dbname, int nrecs));
EXTERN int                SdbBsave _ANSI_ARGS_((Tcl_Interp *interp,
                              SdbFile *sdbPtr, char *filename));
EXTERN int                SdbBload _ANSI_ARGS_((Tcl_Interp *interp,
                              SdbFile *sdbPtr, char *filename));
EXTERN double             SdbPutNumber _ANSI_ARGS_((SdbFile *sdbPtr,
                              SdbVariable *varPtr, double value, 
			      int position));
EXTERN char *             SdbPutCode _ANSI_ARGS_((SdbFile *sdbPtr,
                              SdbVariable *varPtr, char *code, int position));
EXTERN SdbTextFile *      SdbTextOpen _ANSI_ARGS_((Tcl_Interp *interp,
                              char *string, int size));
EXTERN void               SdbTextClose _ANSI_ARGS_((SdbTextFile *file));
EXTERN int                SdbTextGet _ANSI_ARGS_((SdbTextFile *file));
EXTERN int                SdbTextReposition _ANSI_ARGS_((SdbTextFile *file,
                              char *field, int new_pos));
EXTERN SdbNode *          SdbGetNode _ANSI_ARGS_((void));
EXTERN SdbNode *          SdbFreeNode _ANSI_ARGS_((SdbNode *tree));
EXTERN SdbNode *          SdbParse _ANSI_ARGS_((Tcl_Interp *interp,
                              SdbFile *dbPtr, char *string));
EXTERN SdbNode *          SdbOptimize _ANSI_ARGS_((SdbFile *, SdbNode *));
EXTERN int                SdbSnapshot _ANSI_ARGS_((Tcl_Interp *interp, 
			      SdbFile *sdbPtr, char *fileName));
EXTERN int                SdbErrorCheck _ANSI_ARGS_((SdbFile *dpPtr));
EXTERN void               SdbFreshenUp _ANSI_ARGS_((SdbFile *dbPtr, 
				       SdbVariable *varPtr));

#endif
